/* Example program to demonstrate the scheduling of tasks under Mint.
 * This simulates a simple infinite-size cache, using a write-invalidate
 * protocol. Bus contention is modelled by retrying bus requests if
 * the bus is busy.
 */

#include <stdio.h>
#include "mint.h"
#include "hash.h"
#include "bit.h"

#define USAGE "\nsimulator options: \n\
\t[-b size]    (cache block size in bytes)\n\
\t[-o file]    (output file for statistics)\n\
"

/* define a structure to hold information about a cache block */
typedef struct info_t {
    unsigned int	blocknum;
    struct info_t	*next;
    bitvector_t		present;
    bitvector_t		shared;
    bitvector_t		dirty;
} info_t, *info_ptr;

#define ENTRY_SIZE sizeof(struct info_t)

/* hash table size must be a power of two */
#define HASH_TABLE_SIZE (64 * 1024)

/* Cache is a pointer to the hash table data structure for storing
 * cache blocks.
 */
hash_tab_ptr Cache;

/* Outfile is the name of the output file for writing statistics */
static char *Outfile;

/* Fd_out is the file descriptor for the output file. If no output file
 * was specified, it defaults to stderr.
 */
static FILE *Fd_out;

/* Blocksize is the cache block size in bytes */
int Blocksize;
#define DEFAULT_BLOCKSIZE 32

/* Blockshift is the log (base 2) of Blocksize. It is used to convert
 * addresses to cache block numbers.
 */
int Blockshift;

/* for collecting statistics about events */
struct global_stats {
    int read_hit;
    int read_miss;
    int write_hit;
    int write_miss;
} Data;

/* Function prototypes */
/* These functions are called when a bus operation completes */
int read_done(task_ptr ptask);
int excl_done(task_ptr ptask);
int inv_done(task_ptr ptask);

/* These functions are called if the bus was busy and the request
 * must be retried. They are scheduled to run at the next time
 * the bus is free.
 */
int read_retry(task_ptr ptask);
int excl_retry(task_ptr ptask);
int inv_retry(task_ptr ptask);

/* sim_init() is called by Mint once at the beginning of the simulation.
 * The return value is the index in the argv[] array of the next unparsed
 * argument.
 */
sim_init(int argc, char **argv)
{
    int next_arg;
    int exp, logsize;

    next_arg = sim_parse_args(argc, argv);
    /* If an output file was given on the command line,
     * then open the file for writing.
     */
    if (Outfile) {
        Fd_out = fopen(Outfile, "w");
        if (Fd_out == NULL) {
            perror(Outfile);
            fprintf(stderr, USAGE);
            exit(1);
        }
    } else
        Fd_out = stderr;
    /* Change Mint's output to go to the same file. */
    Mint_output = Fd_out;

    Cache = hash_init(HASH_TABLE_SIZE, ENTRY_SIZE);

    /* compute the log (base 2) of Blocksize */
    for (logsize = 0, exp = 1; exp < Blocksize; logsize++)
        exp *= 2;
    Blockshift = logsize;

    return next_arg;
}

/* sim_usage() is called by Mint if there is an error parsing the arguments.
 */
void
sim_usage(int argc, char **argv)
{
    fprintf(stderr, USAGE);
}

/* parse_args() parses the command line arguments and returns the index
 * of the next unparsed argument.
 */
sim_parse_args(int argc, char **argv)
{
    extern char *optarg;
    extern int optind;
    int errflag = 0;
    int c;

    Blocksize = DEFAULT_BLOCKSIZE;
    while ((c = getopt(argc, argv, "b:o:")) != -1)
        switch (c) {
          case 'b':
            Blocksize = strtol(optarg, NULL, 0);
            break;
          case 'o':
            Outfile = optarg;
            break;
          default:
            errflag = 1;
            break;
        }
    if (errflag)
        fatal(USAGE);

    return optind;
}

/*
 * sim_read() is called on every read reference.
 * sim_read() returns:
 *     T_FREE:  if read is not complete
 *     T_ADVANCE   :  if read is complete
 */
sim_read(task_ptr ptask)
{
    int pid, addr, blocknum;
    event_ptr pevent;
    task_ptr newtask;
    info_ptr pinfo;
    
    pid = ptask->pid;
    pevent = ptask->pevent;
    addr = pevent->paddr;

    /* shift the addr to produce a cache block number */
    blocknum = addr >> Blockshift;

    /* lookup the entry for the cache block containing address addr */
    pinfo = (info_ptr) hash_insert(Cache, blocknum);

    /* if the block is present, it's a read hit */
    if (ISVALID(pinfo->present, pid)) {
        Data.read_hit++;
        return T_ADVANCE;
    }

    /* block is not present, issue a bus read request */
    issue_bus_request(ptask, read_retry, read_done, pinfo);

    Data.read_miss++;
    return T_FREE;
}

/*
 * sim_write() is called on every write reference.
 * sim_write() returns:
 *     T_FREE:  if write is not complete
 *     T_ADVANCE   :  if write is complete
 */
sim_write(task_ptr ptask)
{
    int pid, addr, blocknum;
    event_ptr pevent;
    task_ptr newtask;
    info_ptr pinfo;
    
    pid = ptask->pid;
    pevent = ptask->pevent;
    addr = pevent->paddr;

    /* shift the addr to produce a cache block number */
    blocknum = addr >> Blockshift;

    /* lookup the entry for the cache block containing address addr */
    pinfo = (info_ptr) hash_insert(Cache, blocknum);

    /* if the block is present, it's a write hit */
    if (ISVALID(pinfo->present, pid)) {
        Data.write_hit++;
        if (ISVALID(pinfo->shared, pid)) {
            /* block is shared, issue an invalidation */
            issue_bus_request(ptask, inv_retry, inv_done, pinfo);
            return T_FREE;
        } else {
            /* block is not shared, set the dirty bit */
            SETVALID(pinfo->dirty, pid);
            return T_ADVANCE;
        }
    }

    /* block is not present, issue a bus exclusive-read request */
    issue_bus_request(ptask, excl_retry, excl_done, pinfo);

    Data.write_miss++;
    return T_FREE;
}

/* This function is called when a read completes. */
read_done(task_ptr ptask)
{
    info_ptr pinfo;
    int pid;
    
    /* retrieve the process id and the pinfo pointer from this task */
    pid = ptask->pid;
    pinfo = (info_ptr) ptask->uptr;

    /* make this block present in pid's cache */
    SETVALID(pinfo->present, pid);

    /* clear the dirty bits for everyone */
    CLRALLBITS(pinfo->dirty);

    /* thanks to Karl.Westerholz@zfe.siemens.de for a bug fix here */
    /* if anyone else has a copy then this block is known to be shared */
    if (!ISEXCL(pinfo->present, pid))
        SETEQUAL(pinfo->shared, pinfo->present);
    else
        CLRALLBITS(pinfo->shared);

    return T_ADVANCE;
}

/* This function is called when an exclusive read completes. */
excl_done(task_ptr ptask)
{
    info_ptr pinfo;
    int pid;
    
    /* retrieve the process id and the pinfo pointer from this task */
    pid = ptask->pid;
    pinfo = (info_ptr) ptask->uptr;

    /* make this block present only in pid's cache */
    SETEXCL(pinfo->present, pid);

    /* clear the dirty bits for everyone */
    CLRALLBITS(pinfo->dirty);

    /* this block is not shared so clear the shared bits */
    CLRALLBITS(pinfo->shared);

    return T_ADVANCE;
}

/* This function is called when an invalidate completes. */
inv_done(task_ptr ptask)
{
    info_ptr pinfo;
    int pid;

    /* retrieve the process id and the pinfo pointer from this task */
    pid = ptask->pid;
    pinfo = (info_ptr) ptask->uptr;

    /* make this block present only in pid's cache */
    SETEXCL(pinfo->present, pid);

    /* set my dirty bit, clear everyone else's dirty bit */
    SETEXCL(pinfo->dirty, pid);

    /* this block is not shared so clear the shared bits */
    CLRALLBITS(pinfo->shared);

    return T_ADVANCE;
}

/* This function is called if the read request could not be handled
 * immediately because the bus was busy. This function is scheduled by
 * the bus module to execute at the next time the bus becomes free.
 * Since nothing can happen to an invalid block, we simply re-issue
 * the read request.
 */
read_retry(task_ptr ptask)
{
    info_ptr pinfo;

    /* retrieve the pinfo pointer from this task */
    pinfo = (info_ptr) ptask->uptr;

    /* retry the read request */
    issue_bus_request(ptask, read_retry, read_done, pinfo);
    return T_FREE;
}

/* This function is called if the exclusive-read request could not be handled
 * immediately because the bus was busy. This function is scheduled by
 * the bus module to execute at the next time the bus becomes free.
 * Since nothing can happen to an invalid block, we simply re-issue
 * the exclusive-read request.
 */
excl_retry(task_ptr ptask)
{
    info_ptr pinfo;

    /* retrieve the pinfo pointer from this task */
    pinfo = (info_ptr) ptask->uptr;

    /* retry the exclusive-read request */
    issue_bus_request(ptask, excl_retry, excl_done, pinfo);
    return T_FREE;
}

/* This function is called if the invalidation request could not be handled
 * immediately because the bus was busy. This function is scheduled by
 * the bus module to execute at the next time the bus becomes free. Since
 * this block may have been invalidated by some other processor while we
 * were waiting for the bus, we must check if the block is still valid.
 * If the block is still valid, we retry the invalidation (and keep
 * trying until we either get the bus or some other processor invalidates
 * this block). If the block is no longer valid, we need to treat this as
 * a write miss, so we issue an exclusive-read request.
 */
inv_retry(task_ptr ptask)
{
    info_ptr pinfo;
    int pid;

    /* retrieve the process id and the pinfo pointer from this task */
    pid = ptask->pid;
    pinfo = (info_ptr) ptask->uptr;

    /* if this block is still present, then re-issue the invalidation request */
    if (ISVALID(pinfo->present, pid)) {
        issue_bus_request(ptask, inv_retry, inv_done, pinfo);
        return T_FREE;
    }

    /* this block is no longer present, so exclusive-read the cache block */
    issue_bus_request(ptask, excl_retry, excl_done, pinfo);
    return T_FREE;
}

/* This function is called at the end of the simulation to print
 * statistics to a file.
 */
void
sim_done(unsigned long long elapsed, unsigned long long cpu_time)
{
    double total_reads, read_hit_per, read_miss_per;
    double total_writes, write_hit_per, write_miss_per;

    total_reads = Data.read_hit + Data.read_miss;
    read_hit_per = 100.0 * Data.read_hit / total_reads;
    read_miss_per = 100.0 * Data.read_miss / total_reads;

    total_writes = Data.write_hit + Data.write_miss;
    write_hit_per = 100.0 * Data.write_hit / total_writes;
    write_miss_per = 100.0 * Data.write_miss / total_writes;

    fprintf(Fd_out, " Read hits    %%   Read misses    %%\n");
    fprintf(Fd_out, " %9u %5.1f    %9u %5.1f\n",
            Data.read_hit, read_hit_per, Data.read_miss, read_miss_per);
    fprintf(Fd_out, "Write hits    %%  Write misses    %%\n");
    fprintf(Fd_out, " %9u %5.1f    %9u %5.1f\n",
            Data.write_hit, write_hit_per, Data.write_miss, write_miss_per);
}
