#include "mint.h"

#define USAGE "\nsimulator options: [-r] [-w]\n"

/* define a type for memory reference stats: shared and private */
typedef struct memref_s {
    int prv;
    int shr;
} memref_t;

/* for collecting statistics about events */
struct global_stats {
    memref_t nreads;
    memref_t nwrites;
    mint_time_t first_fork_time;
    mint_time_t first_fork_cpu_time;
    mint_time_t last_fork_time;
    mint_time_t last_fork_cpu_time;
    mint_time_t start_time[MAXPROC];
    mint_time_t duration;
    int instr;
    int forks;
    int sprocs;
    int dones;
    int blocks;
    int unblocks;
    int lock_att;
    int lock_acq;
    int lock_rel;
    int barrier_att;
    int barrier_acq;
    int psema_att;
    int psema_acq;
    int vsemas;
    int user;
    int sync;
    int function;
    int other;
} Data;

sim_init(int argc, char **argv)
{
    extern char *optarg;
    extern int optind;
    int c;

    while ((c = getopt(argc, argv, "rw")) != -1)
        switch (c) {
          case 'r':
            /* define this to enable primary cache read hit optimization */
            Use_prmcache_read = 1;
            break;
          case 'w':
            /* define this to enable primary cache write hit optimization */
            Use_prmcache_write = 1;
            break;
          default:
            fatal(USAGE);
        }
    return optind;
}

void
sim_usage(int argc, char **argv)
{
    fprintf(stderr, USAGE);
}

/* for fast processing of reads and writes, define these routines */
prmcache_read(event_ptr pevent)
{
    count_events(pevent);
    /* if the address is private, then do not force a reschedule */
    if ((pevent->type & E_SHARED) == 0)
        return 0;

    /* the address was shared, so force a reschedule */
    return 1;
}

prmcache_write(event_ptr pevent)
{
    count_events(pevent);
    if ((pevent->type & E_SHARED) == 0)
        return 0;
    return 1;
}

count_events(event_ptr pevent)
{
    int pid, type;
    
    pid = pevent->pid;
    type = pevent->type;
    
#ifdef DEBUG_VERBOSE
    printf("p%d event type = 0x%03x, time = %.0f vaddr = 0x%x, paddr = 0x%x, *value = 0x%x, pevent = 0x%x\n",
           pevent->pid, type, (double) pevent->time,
           pevent->vaddr, pevent->paddr,
           *pevent->pvalue, pevent->pevent);
#endif
    
    /* count each type of memory access or other event */
    switch (type) {
      case E_INSTR:
        Data.instr++;
        break;
      case E_READB:
      case E_READBS:
      case E_READH:
      case E_READHS:
      case E_READW:
      case E_LWL:
      case E_LWR:
      case E_FREADW:
      case E_FREADD:
        Data.nreads.prv++;
        break;
        
      case E_SHR_READB:
      case E_SHR_READBS:
      case E_SHR_READH:
      case E_SHR_READHS:
      case E_SHR_READW:
      case E_SHR_LWL:
      case E_SHR_LWR:
      case E_SHR_FREADW:
      case E_SHR_FREADD:
        Data.nreads.shr++;
        break;
        
      case E_WRITEB:
      case E_WRITEH:
      case E_WRITEW:
      case E_SWL:
      case E_SWR:
      case E_FWRITEW:
      case E_FWRITED:
        Data.nwrites.prv++;
        break;
        
      case E_SHR_WRITEB:
      case E_SHR_WRITEH:
      case E_SHR_WRITEW:
      case E_SHR_SWL:
      case E_SHR_SWR:
      case E_SHR_FWRITEW:
      case E_SHR_FWRITED:
        Data.nwrites.shr++;
        break;
        
      case E_FORK:
        Data.forks++;
        break;
        
      case E_FUNCTION:
        Data.function++;
        break;
        
      case E_DONE:
        Data.dones++;
        break;
        
      case E_BLOCK:
        Data.blocks++;
        break;
        
      case E_UNBLOCK:
        Data.unblocks++;
        Data.duration += pevent->duration;
        break;
        
      case E_BARRIER_ATTEMPT:
        Data.barrier_att++;
        break;
        
      case E_BARRIER_ACQUIRE:
        Data.barrier_acq++;
        break;
        
      case E_PSEMA_ATTEMPT:
        Data.psema_att++;
        break;
        
      case E_PSEMA_ACQUIRE:
        Data.psema_acq++;
        break;
        
      case E_VSEMA:
        Data.vsemas++;
        break;
        
      case E_LOCK_ATTEMPT:
        Data.lock_att++;
        break;
        
      case E_LOCK_ACQUIRE:
        Data.lock_acq++;
        break;
        
      case E_LOCK_RELEASE:
        Data.lock_rel++;
        break;
        
      case E_USER:
        Data.user++;
        break;

      case E_SPROC:
        Data.sprocs++;
        break;
        
      default:
        Data.other++;
#ifdef DEBUG_VERBOSE
        printf("p%d event type = 0x%x, time = %.0f vaddr = 0x%x, paddr = 0x%x, *value = 0x%x, pevent = 0x%x\n",
               pevent->pid, type, pevent->time,
               pevent->vaddr, pevent->paddr,
               *pevent->pvalue, pevent->pevent);
#endif
        break;
    }
    return T_ADVANCE;
}

void
sim_done(unsigned long long elapsed, unsigned long long cpu_time)
{
    int total_rd, total_wt, total_priv, total_shar;

    total_priv = Data.nreads.prv + Data.nwrites.prv;
    total_shar = Data.nreads.shr + Data.nwrites.shr;
    total_rd = Data.nreads.prv + Data.nreads.shr;
    total_wt = Data.nwrites.prv + Data.nwrites.shr;

    printf("              Reads     Writes      Total\n");
    printf("Private: %10d %10d %10d\n",
           Data.nreads.prv, Data.nwrites.prv, total_priv);
    printf("Shared:  %10d %10d %10d\n",
           Data.nreads.shr, Data.nwrites.shr, total_shar);
    printf("Total:   %10d %10d %10d\n",
           total_rd, total_wt, total_rd + total_wt);
    printf("Other instructions: %d\n", Data.instr);
    printf("forks: %d\n", Data.forks);
    printf("sprocs: %d\n", Data.sprocs);
    printf("dones: %d, blocks: %d, unblocks: %d\n",
           Data.dones, Data.blocks, Data.unblocks);
    printf("lock_att: %d, lock_acq: %d, lock_rel: %d\n",
           Data.lock_att, Data.lock_acq, Data.lock_rel);
    printf("barrier_att: %d, barrier_acq: %d\n",
           Data.barrier_att, Data.barrier_acq);
    printf("psema_att: %d, psema_acq: %d, vsemas: %d\n",
           Data.psema_att, Data.psema_acq, Data.vsemas);
    printf("user: %d, sync: %d\n", Data.user, Data.sync);
    printf("trapped functions: %d\n", Data.function);
    printf("time spent blocked: %.0f\n", (double) Data.duration);
    printf("Other events: %d\n", Data.other);

    printf("\nFirst fork at time %.0f, cpu time used: %.0f\n",
	   Data.first_fork_time, Data.first_fork_cpu_time);
    printf("Last fork at time %.0f, cpu time used: %.0f\n",
	   Data.last_fork_time, Data.last_fork_cpu_time);
    printf("Elapsed parallel cycles (from last fork to end): %.0f\n",
	   elapsed - Data.last_fork_time);
}

sim_read(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_write(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_lock_attempt(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_lock_acquire(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_lock_release(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_barrier_attempt(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_barrier_acquire(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_psema_attempt(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_psema_acquire(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_vsema(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_instr(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

sim_function(task_ptr ptask)
{
    /* printf("function %s called\n", ptask->pevent->fname); */
    return count_events(ptask->pevent);
}

sim_user(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

void
sim_fork(task_ptr ptask)
{
    int pid, child_pid;
    event_ptr pevent;

    pid = ptask->pid;
    if (Data.first_fork_time == 0.0) {
        Data.first_fork_time = ptask->time;
        Data.first_fork_cpu_time = *ptask->pevent->cpu_time;
    }
    Data.last_fork_time = ptask->time;
    Data.last_fork_cpu_time = *ptask->pevent->cpu_time;
    pevent = ptask->pevent->pevent;
    child_pid = pevent->pid;
    Data.start_time[child_pid] = ptask->time;
#if 0
    printf("sim_fork: %d forking %d at time %.1f\n",
           pid, child_pid, ptask->time);
#endif
    count_events(ptask->pevent);
}


sim_terminate(task_ptr ptask)
{
    return count_events(ptask->pevent);
}

void
sim_block(task_ptr ptask)
{
    count_events(ptask->pevent);
}

void
sim_unblock(task_ptr ptask)
{
    count_events(ptask->pevent);
}
