/* 
 * Please read the COPYRIGHT notice included at the top level
 */
/*
 * Definition of the event structure and bit field mnemonics.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __event_h
#define __event_h

#ifdef ULONG_TIME
/* use unsigned long for time values */
typedef unsigned long mint_time_t, *mint_time_ptr;
/* a constant greater than any thread's time value */
#define MAXTIME (~0)

#else
/* use long long for time values */
typedef unsigned long long mint_time_t, *mint_time_ptr;
/* a constant greater than any thread's time value */
#define MAXTIME (~(long long)0)

#endif

#define MAX_BLOCK  108

typedef struct event {
    mint_time_t time;	  /* time at which this event occurred */
    mint_time_ptr cpu_time; /* pointer to accumulated cpu time of this proccess */
    mint_time_t duration; /* time that an unblocked process spent waiting */
    struct event *next;	  /* pointer to next event (used for E_UNBLOCK_LIST) */
    int pid;		  /* process id */
    int type;		  /* type of instruction (see below) */
    int size;             /* size of data accesses in mem. ref. events */
    long utype;		  /* user-defined types */
    long vaddr;		  /* the effective address the traced program sees */
    long paddr;		  /* the real address in memory */
    char value[MAX_BLOCK];  /* value read or written */
    struct event *pevent; /* pointer to another event structure */
    void *sptr;		  /* used by the task scheduler when a task blocks */
    char *fname;	  /* function name if type E_FUNCTION */
    long arg1;		  /* filled in by MINT on E_FUNCTION event */
    long arg2;		  /* filled in by MINT on E_FUNCTION event */
    long arg3;		  /* filled in by MINT on E_FUNCTION event */
    long arg4;		  /* filled in by MINT on E_FUNCTION event */
    long rval;		  /* the return value from malloc() type calls */
} event_t, *event_ptr;

/*
 *
 * The bits in the "type" field:
 *
 *                12 11 10  9  8  7  6  5  4  3  2  1  0
 *               ----------------------------------------
 *               |  |  |  |  |  |  |  |  |  |  |  |  |  |
 *               ----------------------------------------
 *                         Sh Wr Rd Sp
 *
 *  Bits 6,7,8, and 9 are inherited from MINT
 *  In augmint we also use the six bits 13-18 to encode event types
 *
 *  Sp:	  0 = normal	read or write
 *        1 = special
 *
 *  WrRd: 0 = instruction
 *        1 = read
 *        2 = write
 *
 *  Sh:   0 = private	used by back-end
 *        1 = shared	used by back-end
 * 
 */


/* special events for change in process state */
#define E_SPECIAL	0x40

/* read and write operations */
#define E_READ		0x80
#define E_WRITE		0x100
#define E_MEM_REF	(E_READ | E_WRITE)

/* shared references */
#define E_SHARED	0x200

/*event offsets */
#define OFFSET_SHIFT            13
#define OFFSET_MASK             0x3f  /* up to 64 entries in the jump table */
#define TABLE_OFFSET(TYPE)      ((TYPE>>OFFSET_SHIFT) & OFFSET_MASK)

#define E_READ_OFFSET		(1<<OFFSET_SHIFT)
#define E_WRITE_OFFSET		(2<<OFFSET_SHIFT)

#define E_PREFORK		(E_SPECIAL |  3<<OFFSET_SHIFT)
#define E_FORK			(E_SPECIAL |  4<<OFFSET_SHIFT)

#define E_NEW_LOCK              (E_SPECIAL |  5<<OFFSET_SHIFT)
#define E_INIT_LOCK             (E_SPECIAL |  6<<OFFSET_SHIFT)
#define E_LOCK_ATTEMPT		(E_SPECIAL |  7<<OFFSET_SHIFT)
#define E_LOCK_BLOCK		(E_SPECIAL |  8<<OFFSET_SHIFT)
#define E_LOCK_ACQUIRE		(E_SPECIAL |  9<<OFFSET_SHIFT)
#define E_LOCK_RELEASE		(E_SPECIAL | 10<<OFFSET_SHIFT)
#define E_LOCK_UNBLOCK		(E_SPECIAL | 11<<OFFSET_SHIFT)
#define E_FREE_LOCK             (E_SPECIAL | 12<<OFFSET_SHIFT)

#define E_NEW_BARRIER           (E_SPECIAL | 13<<OFFSET_SHIFT)
#define E_INIT_BARRIER          (E_SPECIAL | 14<<OFFSET_SHIFT)
#define E_BARRIER_ATTEMPT	(E_SPECIAL | 15<<OFFSET_SHIFT)
#define E_BARRIER_BLOCK       	(E_SPECIAL | 16<<OFFSET_SHIFT)
#define E_BARRIER_ACQUIRE	(E_SPECIAL | 17<<OFFSET_SHIFT)
#define E_FREE_BARRIER          (E_SPECIAL | 18<<OFFSET_SHIFT)

#define E_NEW_SEMA              (E_SPECIAL | 19<<OFFSET_SHIFT)
#define E_INIT_SEMA             (E_SPECIAL | 20<<OFFSET_SHIFT)
#define E_PSEMA_ATTEMPT		(E_SPECIAL | 21<<OFFSET_SHIFT)
#define E_PSEMA_BLOCK		(E_SPECIAL | 22<<OFFSET_SHIFT)
#define E_PSEMA_ACQUIRE		(E_SPECIAL | 23<<OFFSET_SHIFT)
#define E_VSEMA			(E_SPECIAL | 24<<OFFSET_SHIFT)
#define E_VSEMA_UNBLOCK		(E_SPECIAL | 25<<OFFSET_SHIFT)
#define E_FREE_SEMA             (E_SPECIAL | 26<<OFFSET_SHIFT)

#define E_SHMALLOC              (E_SPECIAL | 27<<OFFSET_SHIFT)
#define E_SHFREE                (E_SPECIAL | 28<<OFFSET_SHIFT)
#define E_MALLOC                (E_SPECIAL | 29<<OFFSET_SHIFT)
#define E_FREE                  (E_SPECIAL | 30<<OFFSET_SHIFT)

#define E_WAIT_FOR_END          (E_SPECIAL | 31<<OFFSET_SHIFT)

#define E_CHECK_TASKS           (E_SPECIAL | 32<<OFFSET_SHIFT)
#define E_CHECK_CHILDREN        (E_SPECIAL | 33<<OFFSET_SHIFT)
#define E_CHECK_PARENT          (E_SPECIAL | 34<<OFFSET_SHIFT)
#define E_DONE_THR              (E_SPECIAL | 35<<OFFSET_SHIFT)

#define E_USER_EVENT            (E_SPECIAL | 36<<OFFSET_SHIFT)

#define E_SYNCH_TIME 		(E_SPECIAL | 37<<OFFSET_SHIFT)

#define E_LOCK_READ 		(E_SPECIAL | 1<<OFFSET_SHIFT)
#define E_LOCK_WRITE 		(E_SPECIAL | 2<<OFFSET_SHIFT)

#endif /* !__event_h */
