/* 
 * Please read the COPYRIGHT notice included at the top level
 */
/* WHEEL_SIZE must be a power of two */
#define WHEEL_SIZE 32768
#define WHEEL_MASK (WHEEL_SIZE - 1)

/* use only the fields we need and pad the size until it is a power of two */
typedef struct wheel_task {
    struct task *next;		/* forward link for task queue */
    struct task *prev;		/* backward link for task queue */
    mint_time_t time;		/* absolute time at which to execute
                                 * this task */
    int priority;		/* larger numbers = higher priority */
#ifdef ULONG_TIME
    int pad0;
#endif
    int pad1;
    int pad2;
    int pad3;
} wheel_task_t, *wheel_task_ptr;

extern wheel_task_ptr Wheel;
extern wheel_task_ptr Curptr;
extern wheel_task_ptr Endptr;
extern mint_time_t Start_time;
extern mint_time_t End_time;
extern task_ptr Future_q;

void wheel_init ();
task_ptr task_extract ();
task_ptr task_min ();
void task_insert (task_ptr ptask);

/*
 * Define macros so that the calls to task_extract() can be inlined inside
 * the main execution loop.
 *
 * Note: the optimization where Curptr was declared to be a local variable
 * instead of a global can no longer be used because the routine
 * task_insert_current() needs to access Curptr.
 *
 * Two different sets of macros are defined, depending on whether the type of
 * "mint_time_t" is a "long long" or an "unsigned long".
 *
 * These macros are not used if DEBUG is defined, since dbx gets confused
 * by multi-line macros, making debugging nearly impossible.
 */

#if defined(DEBUG) || defined(NO_INLINING)
#define INLINE_TASK_EXTRACT(pnew) pnew = task_extract();
#define INLINE_TASK_INSERT(pnew) task_insert(pnew);
#else
#ifdef ULONG_TIME
#define INLINE_TASK_EXTRACT(pnew) \
{ \
    task_ptr tptr, next; \
    unsigned long index, itime; \
     \
    /* search the time wheel first */ \
    pnew = Curptr->next; \
    if (pnew != (task_ptr) Curptr) { \
        INLINE_REMOVE(pnew); \
    } else { \
         \
        while (1) { \
             \
            /*
             * Search the queues until we get to the end. Since we put \
             * a dummy item in the last (unused) queue, we don't need \
             * to check for overflow of the Curptr in each iteration. \
             */ \
            do { \
                Curptr++; \
                pnew = Curptr->next; \
            } while (pnew == (task_ptr) Curptr); \
             \
            /* if we found a task, remove it and return */ \
            if (Curptr != Endptr) { \
                INLINE_REMOVE(pnew); \
                break; \
            } \
             \
            /* wrap around processing */ \
            Curptr = &Wheel[0]; \
            End_time += WHEEL_SIZE; \
            pnew = Future_q->next; \
            if (pnew == Future_q) { \
                pnew = NULL; \
                break; \
            } \
             \
            for ( ; pnew != Future_q; pnew = next) { \
                next = pnew->next; \
                itime = pnew->time; \
                index = itime & WHEEL_MASK; \
                 \
                if (itime <= End_time) { \
                    /* insert in priority order */ \
                    for (tptr = Wheel[index].prev;
                         tptr->priority < pnew->priority; \
                         tptr = tptr->prev) \
                        ; \
                     \
                    INLINE_REMOVE(pnew); \
                    INLINE_INSERT_AFTER(tptr, pnew); \
                } \
            } \
             \
            pnew = Curptr->next; \
            if (pnew != (task_ptr) Curptr) { \
                INLINE_REMOVE(pnew); \
                break; \
            } \
        } \
    } \
}

#define INLINE_TASK_INSERT(ptask) \
{ \
    unsigned long index, itime; \
    task_ptr tptr; \
 \
    itime = ptask->time; \
 \
    if (itime <= End_time) { \
        index = itime & WHEEL_MASK; \
        /* insert in priority order */ \
        for (tptr = Wheel[index].prev; tptr->priority < ptask->priority; \
             tptr = tptr->prev) \
            ; \
 \
        INLINE_INSERT_AFTER(tptr, ptask); \
    } else \
        INLINE_ENQUEUE(Future_q, ptask); \
}

#else /* ULONG_TIME */

#define INLINE_TASK_EXTRACT(pnew) \
{ \
    task_ptr tptr, next; \
    int diff; \
     \
    /* search the time wheel first */ \
    pnew = Curptr->next; \
    if (pnew != (task_ptr) Curptr) { \
        INLINE_REMOVE(pnew); \
    } else { \
         \
        while (1) { \
             \
            /*
             * Search the queues until we get to the end. Since we put \
             * a dummy item in the last (unused) queue, we don't need \
             * to check for overflow of the Curptr in each iteration. \
             */ \
            do { \
                Curptr++; \
                pnew = Curptr->next; \
            } while (pnew == (task_ptr) Curptr); \
             \
            /* if we found a task, remove it and return */ \
            if (Curptr != Endptr) { \
                INLINE_REMOVE(pnew); \
                break; \
            } \
             \
            /* wrap around processing */ \
            Curptr = &Wheel[0]; \
            End_time += WHEEL_SIZE; \
            Start_time += WHEEL_SIZE; \
            pnew = Future_q->next; \
            if (pnew == Future_q) { \
                pnew = NULL; \
                break; \
            } \
             \
            for ( ; pnew != Future_q; pnew = next) { \
                next = pnew->next; \
                diff = pnew->time - Start_time; \
                 \
                if (diff < WHEEL_SIZE) { \
                    /* insert in priority order */ \
                    for (tptr = Wheel[diff].prev;\
                         tptr->priority < pnew->priority; \
                         tptr = tptr->prev) \
                        ; \
                     \
                    INLINE_REMOVE(pnew); \
                    INLINE_INSERT_AFTER(tptr, pnew); \
                } \
            } \
             \
            pnew = Curptr->next; \
            if (pnew != (task_ptr) Curptr) { \
                INLINE_REMOVE(pnew); \
                break; \
            } \
        } \
    } \
}

#define INLINE_TASK_INSERT(ptask) \
{ \
    int index; \
    task_ptr prev, tptr; \
     \
    index = ptask->time - Start_time; \
    if (index < WHEEL_SIZE) { \
        /* insert in priority order */ \
        for (tptr = Wheel[index].prev; tptr->priority < ptask->priority; \
             tptr = tptr->prev) \
            ; \
 \
        INLINE_INSERT_AFTER(tptr, ptask); \
    } else \
        INLINE_ENQUEUE(Future_q, ptask); \
}

#endif /* !ULONG_TIME */
#endif /* !DEBUG */
