/* 
 * Please read the COPYRIGHT notice included at the top level
 */

/*
 * Routines for reading and parsing the text section and managing the
 * memory for an address space.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <time.h>
#include <sys/types.h>

#if defined(USE_SYSV_IPC)
#include <sys/ipc.h>
#include <sys/shm.h>
#elif defined(USE_MMAP)
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#else
#error "Define either USE_SYSV_IPC or USE_MMAP"
#endif

#include <sys/utsname.h>

#ifdef __svr4__
#include <sys/times.h>
#include <limits.h>
#endif

#define MAIN

#include "icode.h"
#include "globals.h"
#include "version.h"

/*
 * Note: the debugger dbx gets line numbers confused because of embedded
 * newlines inside the double quotes in the following macro.
 */
#ifndef DEBUG
#define USAGE \
"\nUsage: %s [front-end options] [-- simulator options]\
 [-- application options]\n\
\n\
front-end options:\n\
	[-C interval]		checkpoint interval\n\
	[-e file]		use file to generate events\n\
	[-k stack_size]		stack size in bytes, default: %d (0x%x)\n\
	[-h heap_size]		heap size in bytes, default: %d (0x%x)\n\
        [-l page=size,blk=size]	lock allocation policy, \
 default: page=%d,blk=%d\n\
	[-p procs]		number of per-process regions, default: %d\n\
	[-q]			quiet mode -- suppress all warning messages\n\
	[-R]			recycle process ids\n\
	[-s shmem_size]		shared memory size, default: %d (0x%x)\n\
	[-u file]		log events to file for Upshot\n\
	[-V]			verify protocol\n\
	[-v]			print version number\n"

#else
#define USAGE "\nUsage: %s [-m mem_size] [-s shmem_size]\
 [and other stuff...] objfile\n"
#endif

static time_t Start_time;
static time_t Finish_time;
static char Start_date[40];
static char Finish_date[40];
static int Mem_size_start;
static int Mem_size_finish;
#ifdef __svr4__
static struct tms Tms_buffer;
#else
static struct rusage Rusage;
#endif
static mint_time_t Total_elapsed_time;
static mint_time_t Total_cpu_time;

static int Argc;
static char **Argv;

/* imported functions */
void init_all_queues ();
void init_main_thread ();
void malloc_share (thread_ptr child, thread_ptr parent);

/* exported functions */
void mint_init (int argc, char **argv, char **envp);
void mint_done ();

/* private functions */
static void save_args (int argc, char **argv);
static void parse_args (int argc, char **argv);
static void copy_argv (int arg_start, int argc, char **argv);
static void mint_stats ();
static void create_addr_space ();
static void usage ();

void
mint_init (int argc, char **argv, char **envp)
{
    int next_arg;
    FILE *fd;

    Start_time = time (NULL);
    Mem_size_start = (int) sbrk (0);

    Mint_output = stderr;
    Simname = argv[0];
    save_args (argc, argv);
    parse_args (argc, argv);

    next_arg = sim_init (argc, argv);

    init_all_queues ();
    create_addr_space ();
    copy_argv (next_arg, argc, argv);
    init_main_thread ();
    ctxsw_fn = ctxsw_intel;
}

static void
usage ()
{
    fprintf (stderr, USAGE, Simname,
	     STACK_SIZE, STACK_SIZE,
	     HEAP_SIZE, HEAP_SIZE,
	     LOCKPAGE_SIZE, LOCKBLK_SIZE,
	     DEF_MAX_NPROCS,
	     SHMEM_SIZE, SHMEM_SIZE);
}

static void
save_args (int argc, char **argv)
{
    int i;

    /* copy the args */
    Argc = argc;
    Argv = (char **) malloc ((argc + 1) * sizeof (char *));
    if (Argv == NULL)
	fatal ("save_args: cannot allocate 0x%x bytes for Argv.\n",
	       (argc + 1) * sizeof (char *));
    for (i = 0; i < argc; i++) {
	Argv[i] = (char *) malloc (strlen (argv[i]) + 1);
	if (Argv[i] == NULL)
	    fatal ("save_args: cannot allocate 0x%x bytes for Argv[%d].\n",
		   strlen (argv[i]) + 1, i);
	strcpy (Argv[i], argv[i]);
    }
    Argv[argc] = NULL;
}

void
mint_done ()
{
    if (Mint_output)
	mint_stats ();

    sim_done (Total_elapsed_time, Total_cpu_time);
}

static void
mint_stats ()
{
    int i;
    unsigned int elapsed, minutes, seconds;
    unsigned int user_sec, user_min, user_usec, sys_sec, sys_min, sys_usec;
    unsigned int total_sec, total_min, total_usec;
    double user_fsec, sys_fsec, total_fsec, cpu_seconds;
    int mem_used;
    char *machine_name;
    struct utsname utsname;

    Finish_time = time (NULL);
#ifdef __svr4__
    times (&Tms_buffer);
#else
    getrusage (RUSAGE_SELF, &Rusage);
#endif

    fprintf (Mint_output, "\nCommand line: ");
    for (i = 0; i < Argc; i++)
	fprintf (Mint_output, " %s", Argv[i]);
    fprintf (Mint_output, "\n");

    strcpy (Start_date, ctime (&Start_time));

    /* remove trailing newline */
    Start_date[strlen (Start_date) - 1] = 0;
    fprintf (Mint_output, "\nStarted:  %s\n", Start_date);
    strcpy (Finish_date, ctime (&Finish_time));

    /* remove trailing newline */
    Finish_date[strlen (Finish_date) - 1] = 0;
    fprintf (Mint_output, "Finished: %s\n", Finish_date);
    elapsed = Finish_time - Start_time;
    minutes = elapsed / 60;
    seconds = elapsed - (minutes * 60);
    if (uname (&utsname) < 0)
	machine_name = "(unknown)";
    else
	machine_name = utsname.nodename;
    fprintf (Mint_output, "Elapsed time for simulation: %u:%02u on %s\n",
	     minutes, seconds, machine_name);

#ifdef __svr4__
    user_sec = Tms_buffer.tms_utime / CLK_TCK;
    user_usec = 0;
#else
    user_sec = Rusage.ru_utime.tv_sec;
    user_usec = Rusage.ru_utime.tv_usec;
#endif
    user_min = user_sec / 60;
    user_sec = user_sec - (user_min * 60);
    user_fsec = user_sec + (double) user_usec / 1000000.0;

#ifdef __svr4__
    sys_sec = Tms_buffer.tms_stime / CLK_TCK;
    sys_usec = 0;
#else
    sys_sec = Rusage.ru_stime.tv_sec;
    sys_usec = Rusage.ru_stime.tv_usec;
#endif
    sys_min = sys_sec / 60;
    sys_sec = sys_sec - (sys_min * 60);
    sys_fsec = sys_sec + (double) sys_usec / 1000000.0;

    total_min = user_min + sys_min;
    total_sec = user_sec + sys_sec;
    total_usec = user_usec + sys_usec;
    if (total_usec >= 1000000) {
	total_sec++;
	total_usec -= 1000000;
    }
    if (total_sec >= 60) {
	total_min++;
	total_sec -= 60;
    }
    total_fsec = total_sec + (double) total_usec / 1000000.0;
    cpu_seconds = total_min * 60 + total_fsec;

#ifdef 0
    fprintf (Mint_output, "CPU time: user: %u:%05.2f, "
             "system: %u:%05.2f total: %u:%05.2f (%.2f sec)\n",
	     user_min, user_fsec, sys_min, sys_fsec, total_min, total_fsec,
	     cpu_seconds);
#endif
    
    Mem_size_finish = (int) sbrk (0);
    mem_used = Mem_size_finish - Mem_size_start;
    mem_used = (mem_used + 1023) / 1024;
    if (Fevents)
	Total_elapsed_time = cur_time;
    else
	Total_elapsed_time = Threads[0].time;
    Total_cpu_time = Threads[0].cpu_time + Threads[0].child_cpu;
    fprintf (Mint_output, "Space used by malloc: %dK\n", mem_used);

    fprintf (Mint_output, "\nElapsed simulated cycles: %.0f      ",
	     (double)Total_elapsed_time);
    /*
     * fprintf(Mint_output, "\nElapsed simulated cycles: %.0f, cpu
     * cycles: %.0f\n", Total_elapsed_time, Total_cpu_time);
     * fprintf(Mint_output, "Processors used = %d, average
     * cpu_time/proc = %.1f\n\n", Maxpid + 1, Total_cpu_time / (Maxpid
     * + 1));
     */
    fprintf (Mint_output, "Processors used = %d\n\n", Maxpid + 1);
}

/* parse the command line arguments */
static void
parse_args (int argc, char **argv)
{
    char *cp, *tmp;
    int c, errflag;
    extern char *optarg;
    extern int optind;

    /* set up the default values */
    Stack_size = STACK_SIZE;
    Heap_size = HEAP_SIZE;
    Shmem_size = SHMEM_SIZE;
    Ckpoint_freq = DEFAULT_CKPOINT_FREQ;
    Lockpage_size = LOCKPAGE_SIZE;
    Lockblk_size = LOCKBLK_SIZE;
    Mint_debug = 0;
    Verify_protocol = 0;
    Locks_block = 1;
    Recycle_threads = 0;
    Max_nprocs = DEF_MAX_NPROCS;

    errflag = 0;

    while ((c = getopt (argc, argv, "C:De:k:h:l:p:qRs:u:Vv")) != -1)
	switch (c) {
	    case 'C':
		Ckpoint_freq = strtol (optarg, NULL, 0);
		break;
	    case 'D':
		Mint_debug = 1;
		break;
	    case 'e':
		Fevents = optarg;
		break;
	    case 'k':
		Stack_size = strtol (optarg, NULL, 0);
		break;
	    case 'h':
		Heap_size = strtol (optarg, NULL, 0);
		break;
	    case 'l':
		cp = optarg;
		/*
		 * Parse the rest painfully since getopt() doesn't
		 * grok strings.
                 *
		 * Example input:
		 *        -l page=0x400,blk=32
		 * There must not be any spaces in the argument to "-l",
		 * or else getopt() will think it is a filename or a
		 * bad option. 
		 * The comma separated options may appear in any order.
		 */
		while (*cp) {
		    if (!strncmp (cp, "page", 4)) {
			cp += 5;
			Lockpage_size = strtol (cp, &tmp, 0);
			cp = tmp;
		    }
		    else if (!strncmp (cp, "blk", 3)) {
			cp += 4;
			Lockblk_size = strtol (cp, &tmp, 0);
			cp = tmp;
		    }
		    else {
			error ("Unknown option (-l `%s')\n", cp);
			errflag = 1;
			break;
		    }
		    if (*cp == ',')
			cp++;
		}
		if (Lockpage_size < 0)
		    Lockpage_size = 0;
		if (Lockblk_size < 0)
		    Lockblk_size = 0;

		/* round up to next power of two */
		if (Lockpage_size)
		    base2roundup (&Lockpage_size);
		if (Lockblk_size)
		    base2roundup (&Lockblk_size);

		/* make page size as least as large as the block size */
		if (Lockpage_size && Lockpage_size < Lockblk_size)
		    Lockpage_size = Lockblk_size;
		break;
	    case 'p':
		Max_nprocs = strtol (optarg, NULL, 0);
		if (Max_nprocs < 1 || Max_nprocs > MAXPROC) {
		    errflag = 1;
		    fprintf (stderr, "Number of processes must be "
                             "in the range: 1 to %d\n",
			     MAXPROC);
		}
		break;
	    case 'q':
		Quiet = 1;
		break;
	    case 'R':
		Recycle_threads = 1;
		break;
	    case 's':
		Shmem_size = strtol (optarg, NULL, 0);
		break;
	    case 'u':
		Upshot_log = optarg;
		break;
	    case 'V':
		Verify_protocol = 1;
		break;
	    case 'v':
		printf ("mint version %s\n", MINT_VERSION);
		exit (0);
		break;
	    default:
		errflag = 1;
		break;
	}

    if (errflag) {
	usage ();
	sim_usage (argc, argv);
	exit (1);
    }
}

/*
 * copy application args onto global appl_argc and appl_argv for the
 * main application thread, e.g. appl_main(appl_argc, appl_argv).
 */
static void
copy_argv (int arg_start, int argc, char **argv)
{
    int i;
    extern int optind;

    /* one more argument for the appl_sim name */
    appl_argc = argc - arg_start + 1;
    appl_argv = (char **) malloc ((appl_argc + 2) * sizeof (char *));
    if (appl_argv == NULL)
	fatal ("copy_args: cannot allocate 0x%x bytes for appl_argv.\n",
	       (appl_argc + 2) * sizeof (char *));
    appl_argv[0] = (char *) malloc (strlen (argv[0]) + 1);
    if (appl_argv[0] == NULL)
	fatal ("copy_args: cannot allocate 0x%x bytes for appl_argv[0].\n",
	       strlen (argv[0]) + 1);
    strcpy (appl_argv[0], argv[0]);
    for (i = 1; i < appl_argc; i++) {
	appl_argv[i] = (char *) malloc (strlen (argv[i + arg_start - 1]) + 1);
	if (appl_argv[i] == NULL)
	    fatal ("copy_args: cannot allocate 0x%x bytes for "
                   "appl_argv[%d].\n",
		   strlen (argv[i + arg_start - 1]) + 1, i);
	strcpy (appl_argv[i], argv[i + arg_start - 1]);
    }
    appl_argv[appl_argc] = NULL;

    /* reset optind to 1 for application */
    optind = 1;
}

/*
 * logbase2() returns the log (base 2) of its argument, rounded up.
 * It also rounds up its argument to the next higher power of 2.
 */
static int
logbase2 (long *pnum)
{
    int logsize, exp;

    for (logsize = 0, exp = 1; exp < *pnum; logsize++)
	exp *= 2;

    /* round pnum up to nearest power of 2 */
    *pnum = exp;
    return logsize;
}

#ifdef USE_SYSV_IPC

allocate_fixed (long addr, long nbytes)
{
    int id, mem_so_far;
    char *real_addr, *shaddr;
    long size2;

    /*
     * Use shared memory segments to allocate memory at a specific
     * address. The mmap() system call would be more appropriate here,
     * but the DECstation doesn't like it.
     */

    /* round nbytes up to the next power of 2 */
    size2 = nbytes;
    logbase2 (&size2);

    /*
     * Try to allocate the whole shared region in one piece. If that fails,
     * try successively smaller sizes by dividing by 2 each time.
     */
    for (; size2; size2 >>= 1) {
	id = shmget (IPC_PRIVATE, size2, 0777);
	if (id != -1)
	    break;
    }
    if (id == -1) {
	perror ("shmget");
	fprintf (stderr, "allocate_fixed: cannot allocate 0x%x bytes.\n",
		 nbytes);
	fprintf (stderr, "Try using fewer processes or "
                 "allocating less memory per process.\n");
	fprintf (stderr, "(Use the options: -h, -p, or -s.)\n");
	usage ();
	exit (1);
    }

    /* map in the first piece */
    shaddr = (char *) addr;
    real_addr = (char *) shmat (id, shaddr, 0);
    if ((int) real_addr == -1 || real_addr != shaddr) {
	shmctl (id, IPC_RMID, 0);
	perror ("shmat");
	fprintf (stderr, "allocate_fixed: cannot attach "
                 "shared memory segment.\n");
	fprintf (stderr, "Try using fewer processes or "
                 "allocating less memory per process.\n");
	fprintf (stderr, "(Use the options: -h, -p, or -s.)\n");
	usage ();
	exit (1);
    }

    /*
     * Delete the segment now so that we don't need to worry about
     * cleaning up later.
     */
    shmctl (id, IPC_RMID, 0);

    /* map all the rest of the pieces contiguously */
    for (mem_so_far = size2; mem_so_far < nbytes; mem_so_far += size2) {
	id = shmget (IPC_PRIVATE, size2, 0777);
	if (id == -1) {
	    perror ("shmget");
	    fprintf (stderr, "allocate_fixed: can only allocate 0x%x "
                     "out of 0x%x bytes.\n",
		     mem_so_far, nbytes);
	    fprintf (stderr, "Try using fewer processes or "
                     "allocating less memory per process.\n");
	    fprintf (stderr, "(Use the options: -h, -p, or -s.)\n");
	    usage ();
	    exit (1);
	}
	shaddr += size2;
	real_addr = (char *) shmat (id, shaddr, 0);
	if ((int) real_addr == -1 || real_addr != shaddr) {
	    perror ("shmat");
	    fprintf (stderr, "allocate_fixed: cannot attach shared "
                     "memory segment.\n");
	    fprintf (stderr, "Try using fewer processes or "
                     "allocating less memory per process.\n");
	    fprintf (stderr, "(Use the options: -h, -p, or -s.)\n");
	    usage ();
	    shmctl (id, IPC_RMID, 0);
	    exit (1);
	}

	/*
	 * Delete the segment now so that we don't need to worry about
	 * cleaning up later.
	 */
	shmctl (id, IPC_RMID, 0);
    }
}

#endif /* USE_SYSV_IPC */

#ifdef USE_MMAP

/* Allocate memory at a fixed address -- an alternative policy using
 * `mmap' by Richard W.M. Jones. This is more appropriate if your system
 * supports this call, and most do nowadays. In particular, Linux 2.0.0
 * croaks on SYSV IPC even when compiled into the kernel. Why, I don't
 * know.
 * -	rwmj@doc.ic.ac.uk
 */

allocate_fixed (long addr, long nbytes)
{
  static int fd = -1;
  caddr_t real_addr;

  if (fd == -1)
    {
      fd = open ("/dev/zero", O_RDWR);
      if (fd < 0)
	{
	  perror ("open /dev/zero");
	  fprintf (stderr,
		   "allocate_fixed: Failed opening /dev/zero.\n"
		   "  Try using USE_SYSV_IPC instead of USE_MMAP.\n");
	  usage ();
	  exit (1);
	}
    }
  real_addr = mmap ((caddr_t) addr,
		    nbytes,
		    PROT_READ | PROT_WRITE,
		    MAP_PRIVATE | MAP_FILE | MAP_FIXED,
		    fd,
		    0);

  if (real_addr == (caddr_t) -1)
    {
      perror ("mmap");
      fprintf (stderr,
	       "allocate_fixed: Failed mapping memory at a fixed address.\n"
	       "  Try using USE_SYSV_IPC instead of USE_MMAP.\n");
      usage ();
      exit (1);
    }
}

#endif /* USE_MMAP */

/*
 * Allocate memory for the shared memory region and all the per-process
 * private memory regions and initialize the address space for the main
 * process.
 */
static void
create_addr_space ()
{
    int i, size;
    long *addr;
    thread_ptr pthread;
    unsigned long total_size;

    /* Only the stack is allocated in the per thread space */
    Mem_size = Stack_size;

    /* round Mem_size up to the next larger page size */
    Mem_size = (Mem_size + M_ALIGN - 1) & ~(M_ALIGN - 1);

    pthread = &Threads[0];

    /*
     * Allocate all the private memory regions up front.
     * Assign each region to a fixed location so that different
     * simulation runs all use the same starting address for private
     * memory.
     */
    Private_start = PRIVATE_START;
    size = Max_nprocs * Mem_size;
    allocate_fixed (Private_start, size);

    pthread->addr_space = Private_start;

    /* initialize the heap allocator */
    Heap_start = HEAP_START;
    allocate_fixed (Heap_start, Heap_size);
    malloc_init (Heap_start, Heap_size);

    /* allocate the shared memory region */
    Shmem_start = SHARED_START;
    Shmem_end = Shmem_start;
    if (Shmem_size)
	allocate_fixed (Shmem_start, Shmem_size);

    /* allocate a page for locks if necessary */
    if (Shmem_size && Lockpage_size) {
	Lkmem_start = Shmem_end;
	Lkmem_end = Lkmem_start;
	Shmem_end += Lockpage_size;
    }
    else {
	/* initialize the variables so that it appears the space is used up */
	Lkmem_start = 0;
	Lkmem_end = Lockpage_size;
    }
    addr = (long *) pthread->addr_space;

    /* point the sp to the top of the allocated space */
    /* (The stack grows down toward lower memory addresses.) */
    pthread->reg[29] = pthread->addr_space + Mem_size;

    /* Change the sp so that mapping will work for sp-relative addresses. */
    pthread->reg[29] = PUNMAP (pthread->reg[29]);
}

/* Initialize child's address space */
void
init_child_addr_space (thread_ptr parent, thread_ptr child)
{
    int i, pid;
    unsigned int maxdepth;

    /* The address space has already been allocated. */
    pid = child->pid;
    child->addr_space = Private_start + Mem_size * pid;

    /* copy all the registers */
    for (i = 0; i < 32; i++)
	child->reg[i] = parent->reg[i];
    child->lo = parent->lo;
    child->hi = parent->hi;
    for (i = 0; i < 32; i++)
	child->fp[i] = parent->fp[i];
    child->fcr0 = parent->fcr0;
    child->fcr31 = parent->fcr31;

    /* point the sp to the top of the allocated space */
    /* (The stack grows down toward lower memory addresses.) */
    child->reg[29] = child->addr_space + Mem_size;

    /* Change the sp so that mapping will work for sp-relative addresses. */
    child->reg[29] = PUNMAP (child->reg[29]);
}
