/* 
 * Please read the COPYRIGHT notice included at the top level
 */

/*
 * Routines for implementing parallel constructs such as fork(), locks,
 * and semaphores. Includes all the functions that are substituted for
 * their counterparts in the traced object file.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#ifdef sgi
#include <sys/syssgi.h>
#include <sys/sysmp.h>
#endif
#ifndef DECSTATION
#ifndef linux
#include <stropts.h>
#endif
#endif

#include "icode.h"
#include "globals.h"

void sleep_thr (thread_ptr pthread);
void wakeup_all (qnode_ptr q, thread_ptr pthread, task_ptr ptask, int cycles);
void init_thread (thread_ptr pthread, event_ptr pevent);

OP (event_error)
{
    fatal ("unsupported event offset 0x%x\n", pthread->pevent->type);
}

static char *
shmem_alloc (int size)
{
    /* round size up to a double word boundary */
    size = ROUNDUP (size, 8);

    /* remove this space from the shared memory area */
    Shmem_end += size;

    /* make sure there is enough space */
    if (Shmem_end > Shmem_start + Shmem_size)
	fatal ("shmem_alloc: not enough shared memory"
               " (requested size = 0x%x, used = 0x%x)\n",
	       size, Shmem_end - size - Shmem_start);

    return (char *) (Shmem_end - size);
}

/*
 * This routine is called internally by MINT to allocate a barrier,
 * lock, or semaphore.
 */
static char *
synch_alloc (int size)
{
    /* round size up to a double word boundary */
    size = ROUNDUP (size, 8);

    /*
     * If the Lockblk_size is non-zero, then round size up to a
     * block size boundary.
     */
    if (Lockblk_size)
	size = ROUNDUP (size, Lockblk_size);

    /*
     * If Lockpage_size is non-zero, then allocate locks on a page
     * by themselves. Otherwise allocate locks inline with other data.
     */
    if (Lockpage_size) {
	/* allocate space for the lock */
	Lkmem_end += size;

	/* if out of space in this page, then allocate another page */
	if (Lkmem_end > Lkmem_start + Lockpage_size) {
	    if (Lockpage_size < size) {
		/*
		 * If the page size is not large enough to hold one lock,
		 * then increase it to the next larger power of two.
		 */
		Lockpage_size = size;
		base2roundup (&Lockpage_size);
	    }
	    Shmem_end = ROUNDUP (Shmem_end, Lockpage_size);
	    Lkmem_start = (long) shmem_alloc (Lockpage_size);
	    Lkmem_end = Lkmem_start + size;
	}
	return (char *) (Lkmem_end - size);
    }
    /*
     * if we get here, then we are allocating locks inline with other
     * shared data
     */
    return shmem_alloc (size);
}

debug_print_event (task_ptr ptask, thread_ptr pthread, event_ptr
		   pevent, char *st)
{
    printf ("%s:\n", st);
    printf ("   task  : pid %d  \n", ptask->pid);
    printf ("   thread: pid %d  ev 0x%x  tm %f\n",
	    pthread->pid, pthread->pevent, pthread->time);
    printf ("   event : pid %d  tp 0x%x\n", pevent->pid, pevent->type);
}

OP (event_shmalloc)
{
    int size;
    event_ptr pevent;

    pevent = pthread->pevent;
    /* this is the argument of G_MALLOC */
    size = pevent->arg1;

    /* round size up to a double word boundary */
    size = (size + 7) & ~7;

    /* make sure there is enough space */
    if (Shmem_end + size > Shmem_start + Shmem_size)
	fatal ("event_shmalloc: not enough shared memory "
               "(requested size = 0x%x, used = 0x%x)\n",
	       size, Shmem_end - Shmem_start);
    pevent->rval = SUNMAP (Shmem_end);

    /* remove this space from the shared memory area */
    Shmem_end += size;

    ptask->ufunc = sim_shalloc;
    SLEEP (pthread);
}

OP (event_shfree)
{
#ifdef DEBUG_VERBOSE
    printf ("event_shfree()\n");
#endif
    /* share_free() returns 1 on success */
    pthread->pevent->rval = 1;
}

OP (event_prefork)
{
    event_ptr pevent;

    pevent = pthread->pevent;
#ifdef DEBUG_VERBOSE
    printf ("event_prefork( %d bytes )\n", size);
    debug_print_event (ptask, pthread, pthread->pevent, "event_prefork");
#endif
    ptask->ufunc = sim_prefork;
    SLEEP (pthread);
}

OP (event_fork)
{
    int i;
    thread_ptr child;
    int pid;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    printf ("event_fork() time = %.0f\n", pthread->time);
    debug_print_event (ptask, pthread, pthread->pevent, "event_fork");
#endif

    /* allocate a new thread */
    INLINE_DEQUEUE (&Free_q, child);
    if (child == NULL)
	fatal ("event_fork: exceeded thread limit (%d)\n", Max_nprocs);
    if (child->addr_space == 0) {
	NEW_ITEM (Event_free, sizeof (event_t), pevent, "event_fork");
    }
    else
	pevent = child->pevent;
    pid = child->pid;
    if (Maxpid < pid)
	Maxpid = pid;
    Nprocs++;
    if (Nprocs > Max_nprocs)
	fatal ("event_fork: exceeded process limit (%d)\n", Max_nprocs);

    /* initialize the child's address space */
    init_child_addr_space (pthread, child);

    /* init_thread() must be called *after* init_child_addr_space() since
     * init_thread() uses the mapping fields set up by init_child_addr_space().
     */
    init_thread (child, pevent);

    /* the child's start time is the parent's plus the time for the fork */
    pthread->time += FORK_CYCLES;
    child->time = pthread->time;
    if (First_fork_elapsed == 0.0) {
	First_fork_elapsed = child->time;
	pthread->cpu_time += FORK_CYCLES;
	First_fork_cpu = pthread->cpu_time;
    }
    child->parent = pthread;

    /* link in the child to the parent's list of children */
    child->sibling = pthread->youngest;
    pthread->youngest = child;

    /* the first instruction for the child is to run child_envelope */
    child->reg[31] = (int) child_envelope;
    child->ufunc = (PFI) pthread->pevent->arg1;

    /* the return value for the parent is the child's pid */
    pthread->pevent->rval = pid;

    /*
     * task_fork will execute sim_fork and schedule the child's first
     * task
     */
    ptask->ufunc = task_fork;
    pthread->pevent->pevent = pthread->youngest->pevent;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_FORK, pthread->pid, 0, pthread->time);
#endif
}

OP (event_new_barrier)
{
    sema_ptr psema;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_new_barrier");
#endif /* DEBUG_VERBOSE */
    /*
     * Allocate a new semaphore from the shared memory space
     * and store a pointer to it in psema. This should probably
     * use a free list, but semaphores are not likely to be recycled.
     */
    psema = (sema_ptr) synch_alloc (sizeof (sema_t));
#ifdef DEBUG_VERBOSE
    printf ("event_new_barrier (pthread = 0x%x, psema = 0x%x, val = %d)\n",
	    pthread, psema, psema->val);
#endif
    pevent = pthread->pevent;
    pevent->rval = (long) psema;
    ptask->ufunc = sim_lockalloc;
    SLEEP (pthread);
}

OP (event_init_barrier)
{
    sema_ptr psema;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_init_barrier");
#endif /* DEBUG_VERBOSE */
    psema = (sema_ptr) pthread->pevent->arg1;
    if (psema == NULL)
	fatal ("event_init_barrier: bad pointer argument (0x%x)\n", psema);
    psema->val = 0;
    psema->magic = MAGIC_BARRIER;
    INLINE_INIT_Q (&psema->wait_q);
    INLINE_ENQUEUE (&Sema_list, psema);
}

OP (event_barrier_attempt)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent,
                       "event_barrier_attempt");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;
    pevent->paddr = pevent->arg1;
    ptask->ufunc = sim_barrier_attempt;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_BARRIER_ATTEMPT, pthread->pid, 0, pthread->time);
#endif
}

OP (event_barrier_block)
{
    int n;
    sema_ptr psema;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_barrier_block");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    psema = (sema_ptr) pevent->arg1;
    if (psema == NULL || psema->magic != MAGIC_BARRIER)
	fatal ("event_barrier: bad pointer argument (0x%x)\n", psema);
    n = pevent->arg2;
#ifdef DEBUG_VERBOSE
    printf ("event_barrier (pthread = 0x%x, psema = 0x%x, val = %d)\n",
	    pthread, psema, psema->val);
#endif
    if (++psema->val < n) {
	/*
	 * Not every process has arrived at the barrier yet,
	 * so block this process.
	 */
	ptask->ufunc = task_block;
	block_thr (&psema->wait_q, pthread);
    }
    else {
	/* Everyone has arrived, so wake up all blocked processes. */
	wakeup_all (&psema->wait_q, pthread, ptask, BARRIER_CYCLES);
	psema->val = 0;
    }
}

OP (event_barrier_acquire)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent,
                       "event_barrier_acquire");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;	/* address of semaphore */
    pevent->paddr = pevent->arg1;	/* address of semaphore */
    ptask->ufunc = sim_barrier_acquire;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_BARRIER_ACQUIRE, pthread->pid, 0, pthread->time);
#endif
}

OP (event_free_barrier)
{
    sema_ptr psema;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_free_barrier");
#endif /* DEBUG_VERBOSE */
    psema = (sema_ptr) pthread->pevent->arg1;
    if (psema == NULL || psema->magic != MAGIC_BARRIER)
	fatal ("mint_init_barrier: bad pointer argument (0x%x)\n", psema);
    INLINE_REMOVE (psema);
    /* the free list is not currently used */
    FREE_ITEM (Sema_free, psema);
}

OP (event_new_sema)
{
    sema_ptr psema;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_new_sema");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    /*
     * Allocate a new semaphore from the shared memory space
     * and store a pointer to it in psema. This should probably
     * use a free list, but semaphores are not likely to be recycled.
     */
    psema = (sema_ptr) synch_alloc (sizeof (sema_t));
#ifdef DEBUG_VERBOSE
    printf ("event_usnewsema   (pthread = 0x%x, psema = 0x%x, val = %d)\n",
	    pthread, psema, psema->val);
#endif
    pevent->rval = (long) psema;
    ptask->ufunc = sim_lockalloc;
    SLEEP (pthread);
}

OP (event_init_sema)
{
    sema_ptr psema;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_init_sema");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    psema = (sema_ptr) pevent->arg1;
    if (psema == NULL)
	fatal ("mint_usinitsema: bad pointer argument (0x%x)\n", psema);

    psema->magic = MAGIC_SEMA;
    INLINE_INIT_Q (&psema->wait_q);
    INLINE_ENQUEUE (&Sema_list, psema);
    psema->val = pevent->arg2;
#ifdef DEBUG_VERBOSE
    printf ("event_usinitsema  (pthread = 0x%x, psema = 0x%x, val = %d)\n",
	    pthread, psema, psema->val);
#endif
}

OP (event_psema_attempt)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_psema_attempt");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;	/* address of semaphore */
    pevent->paddr = pevent->arg1;	/* address of semaphore */
    ptask->ufunc = sim_psema_attempt;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_PSEMA_ATTEMPT, pthread->pid, 0, pthread->time);
#endif
}

OP (event_psema_block)
{
    sema_ptr psema;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_psema_block");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    psema = (sema_ptr) pevent->arg1;
    if (psema == NULL || psema->magic != MAGIC_SEMA)
	fatal ("event_uspsema: bad pointer argument (0x%x)\n", psema);
#ifdef DEBUG_VERBOSE
    printf ("event_uspsema     (pthread = 0x%x, psema = 0x%x, val = %d)\n",
	    pthread, psema, psema->val);
#endif

    if (--psema->val < 0) {
	ptask->ufunc = task_block;
	block_thr (&psema->wait_q, pthread);
    }
    pevent->rval = 1;
}

OP (event_psema_acquire)
{
    event_ptr pevent;
#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_psema_acquire");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;	/* address of semaphore */
    pevent->paddr = pevent->arg1;	/* address of semaphore */
    ptask->ufunc = sim_psema_acquire;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_PSEMA_ACQUIRE, pthread->pid, 0, pthread->time);
#endif
}

OP (event_vsema)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_vsema");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;	/* address of semaphore */
    pevent->paddr = pevent->arg1;	/* address of semaphore */
    ptask->ufunc = sim_vsema;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_VSEMA, pthread->pid, 0, pthread->time);
#endif
}

OP (event_vsema_unblock)
{
    sema_ptr psema;
    event_ptr pevent;
    thread_ptr pthr;
    mint_time_t duration;

    pevent = pthread->pevent;
    psema = (sema_ptr) pevent->arg1;
    if (psema == NULL || psema->magic != MAGIC_SEMA)
	fatal ("event_vsema_unblock: bad pointer argument (0x%x)\n", psema);
#ifdef DEBUG_VERBOSE
    printf ("event_vsema_unblock   (pthread = 0x%x, psema = 0x%x, val = %d)\n",
	    pthread, psema, psema->val);
    debug_print_event (ptask, pthread, pthread->pevent, "event_vsema_unblock");
#endif /* DEBUG_VERBOSE */

    if (++psema->val <= 0) {
	pthr = unblock_thr ((thread_ptr) & psema->wait_q);
	duration = pthread->time + VSEMA_CYCLES - pthr->time;
	if (duration < 0)
	    fatal ("event_vsema_unblock: negative wait duration "
                   "(%.1f) at time %.0f\n",
		   duration, pthread->time + VSEMA_CYCLES);
	pthr->time += duration;
	/* generate an event for the unblocked thread */
	pevent = pthread->pevent;
	/*
	 * Set the pevent field in the event structure for the unblocked
	 * thread to point to the unblocking thread's event structure so
	 * that sim_block() can figure out who did the unblocking.
	 */
	pthr->pevent->pevent = pevent;
	/*
	 * Set pevent->pevent of the unblocking thread to the pevent
	 * of the unblocked thread
	 */
	pthread->pevent->pevent = pthr->pevent;

	pthread->time += VSEMA_CYCLES;
	ptask->ufunc = task_unblock;
	SLEEP (pthread);
#ifdef LOG_EVENTS
	if (Upshot_log)
	    upshot_log_event (LOG_UNBLOCK, pthr->pid, 0, pthread->time);
#endif
	pthr->pevent->duration = duration;
    }
    pevent->rval = 0;
}

OP (event_free_sema)
{
    sema_ptr psema;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_free_sema");
#endif /* DEBUG_VERBOSE */
    psema = (sema_ptr) pthread->pevent->arg1;
    if (psema == NULL || psema->magic != MAGIC_SEMA)
	fatal ("event_usinitsema: bad pointer argument (0x%x)\n", psema);
    INLINE_REMOVE (psema);
    /* the free list is not currently used */
    FREE_ITEM (Sema_free, psema);
}

OP (event_new_lock)
{
    uslock_ptr plock;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_new_lock");
#endif /* DEBUG_VERBOSE */
    /*
     * Allocate a new lock from the shared memory space
     * and store a pointer to it in plock. This should probably
     * use a free list, but locks are not likely to be recycled.
     */
    plock = (uslock_ptr) synch_alloc (sizeof (uslock_t));
#ifdef DEBUG_VERBOSE
    printf ("event_usnewlock   (pthread = 0x%x, plock = 0x%x, val = %d)\n",
	    pthread, plock, plock->val);
#endif
    pevent = pthread->pevent;
    pevent->rval = (long) plock;
    ptask->ufunc = sim_lockalloc;
    SLEEP (pthread);
}

OP (event_init_lock)
{
    uslock_ptr plock;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_new_lock");
#endif /* DEBUG_VERBOSE */
    /*
     * Allocate a new lock from the shared memory space
     * and store a pointer to it in plock. This should probably
     * use a free list, but locks are not likely to be recycled.
     */
    pevent = pthread->pevent;
    plock = (uslock_ptr) pevent->arg1;
    if (plock == NULL)
	fatal ("event_usinitlock: bad pointer argument (0x%x)\n", plock);
    plock->val = 0;
    plock->pid = -1;
    plock->magic = MAGIC_LOCK;
    INLINE_INIT_Q (&plock->wait_q);
    INLINE_ENQUEUE (&Lock_list, plock);
#ifdef DEBUG_VERBOSE
    printf ("event_lock_init   (pthread = 0x%x, plock = 0x%x, val = %d)\n",
	    pthread, plock, plock->val);
#endif
}

OP (event_lock_attempt)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_lock_attempt");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;	/* address of lock */
    pevent->paddr = pevent->arg1;	/* address of lock */
    ptask->ufunc = sim_lock_attempt;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_LOCK_ATTEMPT, pthread->pid, 0, pthread->time);
#endif
}

OP (event_lock_block)
{
    uslock_ptr plock;
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_lock_block");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    plock = (uslock_ptr) pevent->arg1;
    if (plock == NULL || plock->magic != MAGIC_LOCK)
	fatal ("event_lock_block: bad pointer argument (0x%x)\n", plock);
#ifdef DEBUG_VERBOSE
    printf ("event_lock_block   (pthread = 0x%x, plock = 0x%x, val = %d)\n",
	    pthread, plock, plock->val);
#endif

    if (plock->val == 1) {
	/*
	 * Someone has the lock. Block this process until the lock is
	 * released.
	 * This thread will re-execute this routine when it is awakened.
	 * When it is awakened, it is guaranteed to acquire the lock.
	 */
	ptask->ufunc = task_block;
	block_thr (&plock->wait_q, pthread);
    }
    else {
	/* the lock was free */
	plock->val = 1;
	/* I own the lock */
	plock->pid = pthread->pid;
    }
    pevent->rval = 1;
}

/*
 * Called when a process acquires a lock
 * Return value not used.
 */
OP (event_lock_acquire)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_lock_acquire");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;	/* address of lock */
    pevent->paddr = pevent->arg1;	/* address of lock */
    ptask->ufunc = sim_lock_acquire;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_LOCK_ACQUIRE, pthread->pid, 0, pthread->time);
#endif
}

/*
 * This function is called before releasing a lock.  The lock
 * is released when the back-end allows the processor to continue
 * executing instructions.
 */
OP (event_lock_release)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_lock_release");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    pevent->vaddr = pevent->arg1;	/* address of lock */
    pevent->paddr = pevent->arg1;	/* address of lock */
    ptask->ufunc = sim_lock_release;
    SLEEP (pthread);
#ifdef LOG_EVENTS
    if (Upshot_log)
	upshot_log_event (LOG_LOCK_RELEASE, pthread->pid, 0, pthread->time);
#endif
}

OP (event_lock_unblock)
{
    uslock_ptr plock;
    event_ptr pevent;
    thread_ptr pthr;
    mint_time_t duration;

    pevent = pthread->pevent;
    plock = (uslock_ptr) pevent->arg1;
    if (plock == NULL || plock->magic != MAGIC_LOCK)
	fatal ("event_unlock: bad pointer argument (0x%x)\n", plock);
    if (plock->pid != pthread->pid)
	fatal ("event_unlock: a different thread is unblocking the lock \
than the one that owns it %d  %d\n", plock->pid, pthread->pid);
#ifdef DEBUG_VERBOSE
    printf ("event_unlock (pthread = 0x%x, plock = 0x%x, val = %d)\n",
	    pthread, plock, plock->val);
    debug_print_event (ptask, pthread, pthread->pevent, "event_lock_unblock");
#endif
    /*
     * If there are processes waiting, then unblock one of them.
     */
    if (plock->wait_q.next != &plock->wait_q) {
	pthr = unblock_thr ((thread_ptr) & plock->wait_q);
	duration = pthread->time + UNLOCK_CYCLES - pthr->time;
	if (duration < 0)
	    fatal ("event_unlock: negative wait duration (%.1f) "
                   "at time %.0f\n",
		   duration, pthread->time + UNLOCK_CYCLES);
	pthr->time += duration;
	/*
	 * transfering the ownership of the lock to the unblocked
	 * thread by setting plock->pid to its pid
	 */
	plock->pid = pthr->pid;

	/*
	 * Set the pevent field in the event structure for the unblocked
	 * thread to point to the unblocking thread's event structure so
	 * that sim_block() can figure out who did the unblocking.
	 */
	pthr->pevent->pevent = pevent;
	/*
	 * Set pevent->pevent of the unblocking thread to the pevent
	 * of the unblocked thread
	 */
	pthread->pevent->pevent = pthr->pevent;

	pthread->time += UNLOCK_CYCLES;
	ptask->ufunc = task_unblock;
	SLEEP (pthread);
#ifdef LOG_EVENTS
	if (Upshot_log)
	    upshot_log_event (LOG_UNBLOCK, pthr->pid, 0, pthread->time);
#endif
	pthr->pevent->duration = duration;
    }
    else {
	/* release the lock */
	plock->val = 0;
    }
    pevent->rval = 0;
    /*
     * We already generated the E_LOCK_RELEASE event before executing
     * this function, so just return.
     */
}

OP (event_free_lock)
{
    uslock_ptr plock;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_free_lock");
#endif /* DEBUG_VERBOSE */
    plock = (uslock_ptr) pthread->pevent->arg1;
    if (plock == NULL || plock->magic != MAGIC_LOCK)
	fatal ("event_usfreelock: bad pointer argument (0x%x)\n", plock);
    INLINE_REMOVE (plock);
    /* the free list is not currently used */
    FREE_ITEM (Lock_free, plock);
}

wakeup_thr (thread_ptr pthread)
{
    INLINE_REMOVE (pthread);
    INLINE_ENQUEUE (&Run_q, pthread);
    pthread->runstate = R_RUN;
}

void
wakeup_all (qnode_ptr q, thread_ptr pthread, task_ptr ptask, int cycles)
{
    thread_ptr pthr, tnext;
    mint_time_t duration, wake_time;
    event_ptr pevent;

    pevent = pthread->pevent;
    wake_time = pthread->time + cycles;
    pthr = (thread_ptr) q->next;
    while (pthr != (thread_ptr) q) {
	/* save the pointer to the next thread on the queue */
	tnext = (thread_ptr) pthr->next;
	if (pthr->runstate != R_BLOCK)
	    fatal ("wakeup_all: runstate (%d) not R_BLOCK for thread %d\n",
		   pthr->runstate, pthr->pid);

	/* wake up this thread */
	wakeup_thr (pthr);
	duration = wake_time - pthr->time;
	if (duration < 0)
	    fatal ("wakeup_all: negative wait duration (%.1f) at time %.0f\n",
		   duration, wake_time);
	pthr->time += duration;
	/* if the next field is NULL, then this is first event in list */
	if (pevent->next == NULL) {
	    pevent->next = pthr->pevent;
	    pthr->pevent->next = NULL;
	    /*
	     * Set the pevent field in the event structure for the unblocked
	     * thread to point to the unblocking thread's event structure so
	     * that sim_block() can figure out who did the unblocking.
	     */
	    pthr->pevent->pevent = pevent;
	    pthread->time = wake_time;
	    ptask->ufunc = task_unblock_list;
	    SLEEP (pthread);
	}
	else {
	    /*
	     * If we get here, then this is not the first event in the list,
	     * so add it.
	     */
	    pthr->pevent->next = pevent->next;
	    pevent->next = pthr->pevent;
	}
	pthr->pevent->duration = duration;
	pthr = tnext;
    }
}
